
////////////////////////////////////////////////////////////
//
//  map_compresser.cpp
//

#include <windows.h>
#include "map_compresser.h"

using namespace std;

int LoadIniFile(MapGroup *map_groups, int nMaxCount);
bool Convert_FMF_to_SDK(unsigned char *buf_sdk, const unsigned char *buf_fmf,
						unsigned int size_of_fmf, int vert, int height);


int main(int argc, char *argv[]){

    // n`O[v
    int map_group_count;
    MapGroup map_group[64];
    
    char *filename_in, *filename_header, *filename_out;
    CMPHeader header;
    byte *buf_fmf, *buf_sdk;
    int size_fmf, size_sdk, size_cmp;
	int map_group_num;
	ifstream ifs;
    ofstream ofs;

	//------------------------------------------
    
    cout << "SDK2 Map Compresser\n" << endl;

	// t@C̓ǂݍ
	map_group_count = LoadIniFile(map_group, 64);

    if (argc < 5){
        cout << "Usage: " << argv[0] << " [FMF file] [Header sile] [Output file] [Level number]\n\n""Invalid selection, select from the following:\n" << endl;
		for (int i=0; i<map_group_count; i++){
			cout << i << ": " << map_group[i].name << endl;
		}
        return 0;
    }

	// n`O[vԍ
	map_group_num = atoi(argv[4]);
    if (map_group_num < 0 || map_group_num >= map_group_count){
        cout << "Invalid Selection. Valid numbers are between 0 and " << (map_group_count-1) << "." << endl;
        return -1;
    }
    
    // FMFt@CJ
    filename_in = argv[1];
	ifs.open(filename_in, ios_base::in | ios_base::binary);
	if (ifs.fail()){
		cout << "Failed to open the FMF file." << endl;
		return 0;
	}
    
    // t@CTCY擾
	ifs.seekg(0, ios_base::end);
	size_fmf = ifs.tellg();
	ifs.seekg(0, ios_base::beg);
	size_fmf -= ifs.tellg();
	cout << "FMF file size: " << size_fmf << " bytes." << endl;
    
    // obt@̗pӁEǂݍ݁Et@CN[Y
    buf_fmf = new byte[size_fmf];
    ifs.read((char *)buf_fmf, size_fmf);
    ifs.close();

	// SDK`ɕϊ
	size_sdk = size_fmf - sizeof(FMF_Header);
	buf_sdk = new byte[size_sdk];
	Convert_FMF_to_SDK(buf_sdk, buf_fmf, size_fmf,
		map_group[map_group_num].vertical, map_group[map_group_num].height);
	delete[] buf_fmf;
    
    // kwb_̓ǂݍ
    filename_header = argv[2];
	ifs.open(filename_header, ios_base::in | ios_base::binary);
	if (ifs.fail()){
		cout << "Failed to open the header file." << endl;
		delete[] buf_sdk;
		return 0;
	}
    ifs.read((char *)&header, 0x26);
    ifs.close();
    
    // o͐t@CJ
    filename_out = argv[3];
    ofs.open(filename_out, ios_base::out | ios_base::trunc | ios_base::binary);
	if (ofs.fail()){
		cout << "Failed to save the output file." << endl;
		delete[] buf_sdk;
		return 0;
	}
	size_cmp = -ofs.tellp();

	// kwb_
	ofs.write((const char *)&header, 0x26);
    
    // k
    Compresser *comp = new Compresser(ofs, buf_sdk, size_sdk, header);
    if (comp->compress()){
        cout << "Compression was successful." << endl;
		size_cmp += ofs.tellp();
		cout << "Compressed size: " << size_cmp << "bytes." << endl;
    }else{
        cout << "Compression failed." << endl;
    }
	ofs.close();
	delete comp;
	delete[] buf_sdk;

    return 0;
}



// t@C̓ǂݍ
int LoadIniFile(MapGroup *map_groups, int nMaxCount){

    char IniFileName[260];
	if (!GetModuleFileName(NULL, IniFileName, sizeof IniFileName)) return -1;
	char *last_sep = strrchr(IniFileName, '\\');
	if (last_sep == NULL) return -1;
	strcpy_s(last_sep+1, ((IniFileName + (sizeof IniFileName)) - last_sep) - 1, "map_converter.ini");

    int count = GetPrivateProfileInt("MapGroups", "Count", 15, IniFileName);
    if (count > nMaxCount){
        count = nMaxCount;
    }
    for (int i=0; i<count; i++){
        char str[4];
        sprintf_s(str, sizeof str, "%d", i);
        GetPrivateProfileString("MapGroupName", str, "?",
        map_groups[i].name, sizeof(map_groups[i].name), IniFileName);
        map_groups[i].offset    = GetPrivateProfileInt("MapGroupOffset", str, 0, IniFileName);
        map_groups[i].vertical  = GetPrivateProfileInt("MapGroupVertical", str, 0, IniFileName);
        map_groups[i].height    = GetPrivateProfileInt("MapGroupHeight", str, 16, IniFileName);
    }
    return count;
}



bool Convert_FMF_to_SDK(unsigned char *buf_sdk, const unsigned char *buf_fmf,
						unsigned int size_of_fmf, int vert, int height)
{
	int length = ((size_of_fmf - sizeof(FMF_Header)) / height) >> 1;
	int x_size = vert ? height : length;
	int y_size = vert ? length : height;
	unsigned int chip, hflip, vflip;
	int x, y;
	buf_fmf += sizeof(FMF_Header);
	for (y=0; y<y_size; y++){
		for (x=0; x<x_size; x++){
			chip = ((const unsigned short*)buf_fmf)[x + y*x_size];
			hflip = chip & 0x0010;
			vflip = chip & 0x0020;
			chip = ((chip & 0xff00) >> 4) | (chip & 0x000f);
			chip |= ((hflip!=0) << 14) | ((vflip!=0) << 15);
			((unsigned short*)buf_sdk)[vert ? x + y*height : x*height + y] = chip;
		}
	}
	return true;
}












