
////////////////////////////////////////////////////////////
//
//  map_converter.cpp
//

#undef _DEBUG

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <windows.h>
#include "map_converter.h"

#define BUF_SIZE 65536

int LoadIniFile(MapGroup *map_groups, int nMaxCount);
void show_usage(char *argv_0, MapGroup *map_groups, int map_group_count);
void show_error_expansion(const unsigned char *buf, const ErrorInfo *error);
int expand_data(unsigned char *buf, int buf_size, CMPHeader *cmp_header, FILE *file, ErrorInfo *error);
int convert_to_fmf(FMFHeader *fmf_header, unsigned short *buf_fmf,
    const unsigned char *buf_sdk, unsigned int size, int vert, int height);


int main(int argc, char *argv[]){
    
    // n`O[v
    int map_group_count;
    MapGroup map_group[64];
    
    FILE *file;
    int map_group_num;
    
    CMPHeader cmp_header;
    char buf_layer[BUF_SIZE];
    int size;
    ErrorInfo error;
    
    FMFHeader fmf_header;
    unsigned short *buf_fmf_layer;
    
    //----------------------------------
    
    printf("SDK2 Map Exporter\n\n");
    
    // ݒt@C̓ǂݍ
    map_group_count = LoadIniFile(map_group, 64);

    // ̕\
    if (argc < 5){
        show_usage(argv[0], map_group, map_group_count);
        return 0;
    }

    // n`O[vԍ̎擾
    map_group_num = atoi(argv[4]);
    if (map_group_num < 0 || map_group_num >= map_group_count){
        printf("Invalid Selection. Valid numbers are between 0 and %d.\n", map_group_count-1);
        return -1;
    }
    
    // ROMt@CI[v
    file = fopen(argv[1], "rb");
    if (file == NULL){
        printf("Failed to open the ROM file. \n");
        return -1;
    }
    fseek(file, map_group[map_group_num].offset, SEEK_SET);

    // kwb_[ݒ(PRs[邾BACg͋NȂ)
    fread(&cmp_header, SIZE_OF_CMP_HEADER, 1, file);

    // {f[^̓WJ
    printf("Expanding compressed map data ... \n");
    size = expand_data(buf_layer, sizeof buf_layer, &cmp_header, file, &error);
    if (size == -1){
        fclose(file);
        return -1;
    }
    printf("Current file position: %08X\n", ftell(file));
    printf("Data expanded successfully.\nSaving header file...\n");
    fclose(file);
    
    // wb_t@Co
    file = fopen(argv[3], "wb");
    if (file == NULL){
        printf("Failed to save the header file.\n");
    }else{
        fwrite(&cmp_header, SIZE_OF_CMP_HEADER, 1, file);
        fclose(file);
        printf("Successfully saved header file. \n");
    }
    
    // WJ̃f[^(SDKƓ`)o
    file = fopen("exp_map.dat", "wb");
    if (file == NULL){
        printf("Failed to save uncompressed file.\n");
    }else{
        fwrite(buf_layer, size, 1, file);
        fclose(file);
        printf("Successfully saved uncompressed data.\n");
    }
    
    // FMF`ɕϊ
    buf_fmf_layer = (unsigned short *)malloc(size);
    if (buf_fmf_layer == NULL){
        printf("Form has failed to secure a buffer for FMF.\n");
        return -1;
    }
    if (convert_to_fmf(&fmf_header, buf_fmf_layer, buf_layer, size,
    map_group[map_group_num].vertical, map_group[map_group_num].height) == -1){
        printf("Failed to convert to FMF.\n");
        free(buf_fmf_layer);
        return -1;
    }else{
        printf("Successfully converted to FMF.\n");
    }
    
    // FMFt@CJ
    printf("Saving FMF file...\n");
    file = fopen(argv[2], "wb");
    if (file == NULL){
        printf("Failed to save output FMF file.\n");
        return -1;
    }
    
    // FMFo
    fwrite(&fmf_header, sizeof(FMFHeader), 1, file);    // FMFwb_[
    fwrite(buf_fmf_layer, size, 1, file);               // C[f[^
    free(buf_fmf_layer);
    fclose(file);
    printf("Successfully output FMF file.\n");

    return 0;
}



// t@C̓ǂݍ
int LoadIniFile(MapGroup *map_groups, int nMaxCount){

    char IniFileName[260];
	if (!GetModuleFileName(NULL, IniFileName, sizeof IniFileName)) return -1;
	char *last_sep = strrchr(IniFileName, '\\');
	if (last_sep == NULL) return -1;
	strcpy(last_sep+1, "map_converter.ini");

    int count = GetPrivateProfileInt("MapGroups", "Count", 15, IniFileName);
    if (count > nMaxCount){
        count = nMaxCount;
    }
    for (int i=0; i<count; i++){
        char str[4];
        sprintf(str, "%d", i);
        GetPrivateProfileString("MapGroupName", str, "?",
        map_groups[i].name, sizeof(map_groups[i].name), IniFileName);
        map_groups[i].offset    = GetPrivateProfileInt("MapGroupOffset", str, 0, IniFileName);
        map_groups[i].vertical  = GetPrivateProfileInt("MapGroupVertical", str, 0, IniFileName);
        map_groups[i].height    = GetPrivateProfileInt("MapGroupHeight", str, 16, IniFileName);
    }
    return count;
}


// gp@\
void show_usage(char *argv_0, MapGroup *map_groups, int map_group_count){
    printf("Usage: [ROM file name] [FMF file] [Header file] [Level number]\n\n"
    "Invalid selection, select from the following:\n\n", argv_0);
    for (int i=0; i<map_group_count; i++){
        printf("%d: %s\n", i, map_groups[i].name);
    }
    return;
}


// f[^WJB΃f[^TCYAs-1ԂB
int expand_data(unsigned char *buf, int buf_size, CMPHeader *cmp_header, FILE *file, ErrorInfo *error){
    
    unsigned int cbuf = 0;  // obt@̃JE^
    unsigned int s4 = 0;    // 4bitꂽԂ̂Ƃ1
    unsigned int s4d;       // 4bitꂽԂ̂ƂAߏɓǂݏo4biti[

    unsigned int byte;
    unsigned int sz;
    unsigned int dt;
    unsigned int i;
    
    while (1){

        // 4bitĂ邩ǂŕ
        if (s4){
#ifdef _DEBUG
            printf("p");
#endif
            // ĂƂ : s4d̒lŔf
            switch (s4d){
                case 0x0:       // 0X : ROMXoCgRs[B00̏ꍇIB꒼ȂB
#ifdef _DEBUG
            printf("0 ");
#endif
                    byte = fgetc(file);                         // 1byteǂ
                    if ((sz = (byte & 0xf0) >> 4) == 0){
                        return cbuf;                            // I(4bit͖)
                    }else{
                        if (cbuf + sz > buf_size){              // I[o[t[
                            error->code = ERROR_OVERFLOW;
                            error->cbuf = cbuf;
                            error->type = 0x10;
                            error->param = sz;
                            return -1;
                        }
                        for (i=0; i<sz; i++){
                            dt = fgetc(file);                   // 1byteǂ
                            byte = ((byte&0x0f)<<4)|((dt&0xf0)>>4); // ݂̉4bitƎ̏4bitA
                            buf[cbuf++] = byte;                 // A8bit
                            byte = dt;                          // ւ
                        }
                        s4d = byte & 0x0f;                      // ꒼Ȃ(4bitɎz)
                    }
                    break;
                case 0x1:       // 1XX : XX(8bit)͂B꒼B
#ifdef _DEBUG
            printf("1 ");
#endif
                    if (cbuf + 1 > buf_size){                   // I[o[t[
                        error->code = ERROR_OVERFLOW;
                        error->cbuf = cbuf;
                        error->type = 0x11;
                        error->param = 1;
                        return -1;
                    }
                    buf[cbuf++] = fgetc(file);                  // 1byte(XX)ǂݏ
                    s4=0;                                       // ꒼
                    break;
                case 0x2:       // 2XXYY : XX YY(16bit)͂B꒼B
#ifdef _DEBUG
            printf("2 ");
#endif
                    if (cbuf + 2 > buf_size){                   // I[o[t[
                        error->code = ERROR_OVERFLOW;
                        error->cbuf = cbuf;
                        error->type = 0x12;
                        error->param = 2;
                        return -1;
                    }
                    buf[cbuf++] = fgetc(file);                  // 1byte(XX)ǂݏ
                    buf[cbuf++] = fgetc(file);                  // 1byte(YY)ǂݏ
                    s4=0;                                       // ꒼
                    break;
                case 0x3:       // 3XYY : YY(8bit)(X+3)͂B꒼ȂB
                    dt = fgetc(file);                           // 1byteǂ
#ifdef _DEBUG
            printf("3 ");
#endif
                    sz = ((dt & 0xf0) >> 4) + 3;                // 񐔂̌vZ
                    if (cbuf + sz > buf_size){                  // I[o[t[
                        error->code = ERROR_OVERFLOW;
                        error->cbuf = cbuf;
                        error->type = 0x13;
                        error->param = sz;
                        return -1;
                    }
                    byte = fgetc(file);                         // 1byteǂ
                    dt = ((dt&0x0f) << 4) | ((byte&0xf0) >> 4); // 8bitɌ
                    for (i=0; i<sz; i++) buf[cbuf++] = dt;      // dtsz񏑂
                    s4d = byte & 0x0f;                          // ꒼Ȃ
                    break;
                case 0x4:       // 4X : cmp_header->byte[0] (8bit)(X+3)񏑂݁B꒼ȂB
#ifdef _DEBUG
            printf("4 ");
#endif
                    byte = fgetc(file);                         // 1byteǂ
                    sz = ((byte & 0xf0) >> 4) + 3;              // 񐔂̌vZ
                    if (cbuf + sz > buf_size){                  // I[o[t[
                        error->code = ERROR_OVERFLOW;
                        error->cbuf = cbuf;
                        error->type = 0x14;
                        error->param = sz;
                        return -1;
                    }
                    for (i=0; i<sz; i++) buf[cbuf++] = cmp_header->byte[0];
                    s4d = byte & 0x0f;                          // 8bitf[^sz񏑂݁A꒼Ȃ
                    break;
                case 0x5:       // 5X : cmp_header->byte[1] (8bit)(X+3)񏑂݁B꒼ȂB
#ifdef _DEBUG
            printf("5 ");
#endif
                    byte = fgetc(file);                         // 1byteǂ
                    sz = ((byte & 0xf0) >> 4) + 3;              // 񐔂̌vZ
                    if (cbuf + sz > buf_size){                  // I[o[t[
                        error->code = ERROR_OVERFLOW;
                        error->cbuf = cbuf;
                        error->type = 0x15;
                        error->param = sz;
                        return -1;
                    }
                    for (i=0; i<sz; i++) buf[cbuf++] = cmp_header->byte[1];
                    s4d = byte & 0x0f;                          // 8bitf[^sz񏑂, ꒼Ȃ
                    break;
                case 0x6:       // 6 : cmp_header->word[0] (16bit)1񏑂݁B꒼B
#ifdef _DEBUG
            printf("6 ");
#endif
                    if (cbuf + 2 > buf_size){                   // I[o[t[
                        error->code = ERROR_OVERFLOW;
                        error->cbuf = cbuf;
                        error->type = 0x16;
                        error->param = 2;
                        return -1;
                    }
                    *(unsigned short*)(buf+cbuf) = cmp_header->word[0];
                    cbuf+=2;                                    // 16bitRs[
                    s4=0;                                       // ꒼
                    break;
                case 0x7:       // 7 : cmp_header->byte[2] (8bit)1񏑂݁B꒼B
#ifdef _DEBUG
            printf("7 ");
#endif
                    if (cbuf + 1 > buf_size){                   // I[o[t[
                        error->code = ERROR_OVERFLOW;
                        error->cbuf = cbuf;
                        error->type = 0x17;
                        error->param = 1;
                        return -1;
                    }
                    buf[cbuf++] = cmp_header->byte[2];          // 8bitRs[
                    s4=0;                                       // ꒼
                    break;
                case 0x8:       // 8 : cmp_header->byte[3] (8bit)1񏑂݁B꒼B
#ifdef _DEBUG
            printf("8 ");
#endif
                    if (cbuf + 1 > buf_size){                   // I[o[t[
                        error->code = ERROR_OVERFLOW;
                        error->cbuf = cbuf;
                        error->type = 0x18;
                        error->param = 1;
                        return -1;
                    }
                    buf[cbuf++] = cmp_header->byte[3];          // 8bitRs[
                    s4=0;                                       // ꒼
                    break;
                case 0x9:       // 9X : (X+2)oCgO2oCgRs[B꒼ȂB
#ifdef _DEBUG
            printf("9 ");
#endif
                    if (cbuf + 2 > buf_size){                   // I[o[t[
                        error->code = ERROR_OVERFLOW;
                        error->cbuf = cbuf;
                        error->type = 0x19;
                        error->param = 2;
                        return -1;
                    }
                    byte = fgetc(file);                         // 1byteǂ
                    dt = ((byte & 0xf0) >> 4) + 2;              // ʒuvZ
                    if (cbuf<dt || dt<2){                       // sȈʒu
                        error->code = ERROR_ILG_POS;
                        error->cbuf = cbuf;
                        error->type = 0x19;
                        error->param = dt;
                        return -1;
                    }
                    *(unsigned short*)(buf+cbuf) = *(unsigned short*)(buf+cbuf-dt);
                    cbuf+=2;                                    // 16bitRs[
                    s4d = byte & 0x0f;                          // ꒼Ȃ
                    break;
                case 0xA:       // AXYY : {(X+3)+YY}oCgO(X+3)oCgRs[B꒼ȂB
#ifdef _DEBUG
            printf("A ");
#endif
                    dt = fgetc(file);                           // 1byte
                    sz = ((dt & 0xf0) >> 4) + 3;                // TCYvZ
                    if (cbuf + sz > buf_size){                  // I[o[t[
                        error->code = ERROR_OVERFLOW;
                        error->cbuf = cbuf;
                        error->type = 0x1A;
                        error->param = sz;
                        return -1;
                    }
                    byte = fgetc(file);                         // Ɏ1byte
                    dt = ((dt&0x0f) << 4) | ((byte&0xf0) >> 4); // YY8bitɌ
                    dt += sz;                                   // X+3𑫂Ĉʒum
                    if (cbuf<dt || sz>dt){                      // ʒu܂̓TCYs
                        error->code = ERROR_ILG_POS;
                        error->cbuf = cbuf;
                        error->type = 0x1A;
                        error->param = dt;
                        return -1;
                    }
                    memcpy(buf+cbuf, buf+cbuf-dt, sz);          // Rs[
                    cbuf += sz;
                    s4d = byte & 0x0f;                          // ꒼Ȃ
                    break;
                case 0xB:       // BXYYZ : (YYZ + 103h)oCgO(X+3)oCgRs[B꒼B
#ifdef _DEBUG
            printf("B ");
#endif
                    dt = fgetc(file);                           // 1byte
                    sz = ((dt & 0xf0) >> 4) + 3;                // TCYvZ
                    if (cbuf + sz > buf_size){                  // I[o[t[
                        error->code = ERROR_OVERFLOW;
                        error->cbuf = cbuf;
                        error->type = 0x1B;
                        error->param = sz;
                        return -1;
                    }
                    byte = fgetc(file);                         // Ɏ1byte
                    dt = (((dt&0x0f)<<8) | byte) + 0x103;       // ʒuvZ
                    if (cbuf<dt || sz>dt){                      // ʒu܂̓TCYs
                        error->code = ERROR_ILG_POS;
                        error->cbuf = cbuf;
                        error->type = 0x1B;
                        error->param = dt;
                        return -1;
                    }
                    memcpy(buf+cbuf, buf+cbuf-dt, sz);          // Rs[
                    cbuf += sz;
                    s4 = 0;                                     // ꒼
                    break;
                case 0xC:       // CXYYZZ : YYZZ(BigEndian)oCgO(X+3)oCgRs[B꒼ȂB
#ifdef _DEBUG
            printf("C ");
#endif
                    dt = fgetc(file);                           // 1byte(XY)
                    sz = ((dt & 0xf0) >> 4) + 3;                // TCYvZ
                    if (cbuf + sz > buf_size){                  // I[o[t[
                        error->code = ERROR_OVERFLOW;
                        error->cbuf = cbuf;
                        error->type = 0x1C;
                        error->param = sz;
                        return -1;
                    }
                    byte = fgetc(file);                         // 1byte(YZ)
                    dt = ((dt & 0x0f) << 8) | byte;             // ܂YYZ߂
                    byte = fgetc(file);                         // Ɏ1byte(Z*)
                    dt = (dt << 4) | ((byte & 0xf0) >> 4);      // ʒům(YYZZ)
                    if (cbuf<dt || sz>dt){                      // ʒu܂̓TCYs
                        error->code = ERROR_ILG_POS;
                        error->cbuf = cbuf;
                        error->type = 0x1C;
                        error->param = dt;
                        return -1;
                    }
                    memcpy(buf+cbuf, buf+cbuf-dt, sz);          // Rs[
                    cbuf += sz;
                    s4d = byte & 0x0f;                          // ꒼Ȃ
                    break;
                case 0xD:       // D : 1oCgO1oCgRs[B꒼B
#ifdef _DEBUG
            printf("D ");
#endif
                    if (cbuf + 1 > buf_size){                   // I[o[t[
                        error->code = ERROR_OVERFLOW;
                        error->cbuf = cbuf;
                        error->type = 0x1D;
                        error->param = 1;
                        return -1;
                    }
                    if (cbuf == 0){                             // ʒus
                        error->code = ERROR_ILG_POS;
                        error->cbuf = cbuf;
                        error->type = 0x1D;
                        error->param = 1;
                        return -1;
                    }
                    buf[cbuf] = buf[cbuf-1];                    // 8bitRs[
                    cbuf++; s4=0;                               // ꒼
                    break;
                case 0xE:       // E : 2oCgO2oCgRs[B꒼B
#ifdef _DEBUG
            printf("E ");
#endif
                    if (cbuf + 2 > buf_size){                  // I[o[t[
                        error->code = ERROR_OVERFLOW;
                        error->cbuf = cbuf;
                        error->type = 0x1E;
                        error->param = 2;
                        return -1;
                    }
                    if (cbuf < 2){                              // ʒus
                        error->code = ERROR_ILG_POS;
                        error->cbuf = cbuf;
                        error->type = 0x1E;
                        error->param = 2;
                        return -1;
                    }
                    *(unsigned short*)(buf+cbuf) = *(unsigned short*)(buf+cbuf-2);
                    cbuf+=2; s4=0;                              // 16bitRs[, ꒼
                    break;
                case 0xF:       // FX : cmp_header->word[X+1] (16bit)݁B꒼ȂB
#ifdef _DEBUG
            printf("F ");
#endif
                    if (cbuf + 2 > buf_size){                   // I[o[t[
                        error->code = ERROR_OVERFLOW;
                        error->cbuf = cbuf;
                        error->type = 0x1F;
                        error->param = 2;
                        return -1;
                    }
                    byte = fgetc(file);                         // 1byte
                    dt = ((byte & 0xf0) >> 4) + 1;              // ԍvZ
                    *(unsigned short*)(buf+cbuf) = cmp_header->word[dt];
                    cbuf+=2; s4d = byte & 0x0f;                 // 16bitRs[, ꒼Ȃ
                    break;
            }
        }else{
#ifdef _DEBUG
            printf("n");
#endif
            // ĂȂƂ : ܂1byteǂ݂B̏4bitŔf
            byte = fgetc(file);
            //if (byte == EOF) return -1;
            
            //printf("%x", byte);
            
            switch ((byte & 0xf0) >> 4){
                case 0x0:       // 0X : ROMXoCgRs[B00̏ꍇIBȂB
#ifdef _DEBUG
            printf("0 ");
#endif
                    if ((sz = byte & 0x0f) == 0x0){
                        return cbuf;                            // I
                    }else{
                        if (cbuf + sz > buf_size){              // I[o[t[
                            error->code = ERROR_OVERFLOW;
                            error->cbuf = cbuf;
                            error->type = 0x00;
                            error->param = sz;
                            return -1;
                        }
                        fread(buf+cbuf, sz, 1, file);           // szoCgǂݍ
                        cbuf += sz;                             // ĂȂ܂܃^[Gh
                    }
                    break;
                case 0x1:       // 1XX : XX(8bit)͂BB
#ifdef _DEBUG
            printf("1 ");
#endif
                    if (cbuf + 1 > buf_size){                   // I[o[t[
                        error->code = ERROR_OVERFLOW;
                        error->cbuf = cbuf;
                        error->type = 0x01;
                        error->param = 1;
                        return -1;
                    }
                    dt = (byte & 0x0f) << 4;                    // 4bit
                    byte = fgetc(file);                         // 1byteǂ
                    dt |= (byte & 0xf0) >> 4;                   // 4bit
                    buf[cbuf++] = dt;                           // 8bit
                    s4=1, s4d = byte & 0x0f;                    // 4bit
                    break;
                case 0x2:       // 2XXYY : XX YY(16bit)͂BB
#ifdef _DEBUG
            printf("2 ");
#endif
                    if (cbuf + 2 > buf_size){                   // I[o[t[
                        error->code = ERROR_OVERFLOW;
                        error->cbuf = cbuf;
                        error->type = 0x02;
                        error->param = 2;
                        return -1;
                    }
                    dt = (byte & 0x0f) << 4;                    // 4bit
                    byte = fgetc(file);                         // 1byteǂ
                    dt |= (byte & 0xf0) >> 4;                   // 4bit
                    buf[cbuf++] = dt;                           // 1byteڂ
                    dt = (byte & 0x0f) << 4;                    // 4bit
                    byte = fgetc(file);                         // 1byteǂ
                    dt |= (byte & 0xf0) >> 4;                   // 4bit
                    buf[cbuf++] = dt;                           // 2byteڂ
                    s4=1, s4d = byte & 0x0f;                    // 4bit
                    break;
                case 0x3:       // 3XYY : YY (8bit)(X+3)͂BȂB
#ifdef _DEBUG
            printf("3 ");
#endif
                    sz = (byte & 0x0f) + 3;                     // 񐔌vZ
                    if (cbuf + sz > buf_size){                  // I[o[t[
                        error->code = ERROR_OVERFLOW;
                        error->cbuf = cbuf;
                        error->type = 0x03;
                        error->param = sz;
                        return -1;
                    }
                    byte = fgetc(file);                         // 1byteǂ(YY)
                    for (i=0; i<sz; i++) buf[cbuf++] = byte;    // bytesz񏑂
                    break;
                case 0x4:       // 4X : cmp_header->byte[0] (8bit)(X+3)͂BȂB
#ifdef _DEBUG
            printf("4 ");
#endif
                    sz = (byte & 0x0f) + 3;                     // 񐔌vZ
                    if (cbuf + sz > buf_size){                  // I[o[t[
                        error->code = ERROR_OVERFLOW;
                        error->cbuf = cbuf;
                        error->type = 0x04;
                        error->param = sz;
                        return -1;
                    }
                    for (i=0; i<sz; i++)
                        buf[cbuf++] = cmp_header->byte[0];      // 8bitf[^sz񏑂
                    break;
                case 0x5:       // 5X : cmp_header->byte[1] (8bit)(X+3)͂BȂB
#ifdef _DEBUG
            printf("5 ");
#endif
                    sz = (byte & 0x0f) + 3;                     // 񐔌vZ
                    if (cbuf + sz > buf_size){                  // I[o[t[
                        error->code = ERROR_OVERFLOW;
                        error->cbuf = cbuf;
                        error->type = 0x05;
                        error->param = sz;
                        return -1;
                    }
                    for (i=0; i<sz; i++)
                        buf[cbuf++] = cmp_header->byte[1];      // 8bitf[^sz񏑂
                    break;
                case 0x6:       // 6 : cmp_header->word[0] (16bit)1͂BB
#ifdef _DEBUG
            printf("6 ");
#endif
                    if (cbuf + 2 > buf_size){                   // I[o[t[
                        error->code = ERROR_OVERFLOW;
                        error->cbuf = cbuf;
                        error->type = 0x06;
                        error->param = 2;
                        return -1;
                    }
                    *(unsigned short*)(buf+cbuf) = cmp_header->word[0];
                    cbuf += 2;                                  // 16bitRs[
                    s4=1, s4d = byte & 0x0f;                    // 4bit
                    break;
                case 0x7:       // 7 : cmp_header->byte[2] (8bit)1͂BB
#ifdef _DEBUG
            printf("7 ");
#endif
                    if (cbuf + 1 > buf_size){                   // I[o[t[
                        error->code = ERROR_OVERFLOW;
                        error->cbuf = cbuf;
                        error->type = 0x07;
                        error->param = 1;
                        return -1;
                    }
                    buf[cbuf++] = cmp_header->byte[2];          // ̂܂8bit
                    s4=1, s4d = byte & 0x0f;                    // 4bit
                    break;
                case 0x8:       // 8 : cmp_header->byte[3] (8bit)1͂BB
#ifdef _DEBUG
            printf("8 ");
#endif
                    if (cbuf + 1 > buf_size){                   // I[o[t[
                        error->code = ERROR_OVERFLOW;
                        error->cbuf = cbuf;
                        error->type = 0x08;
                        error->param = 1;
                        return -1;
                    }
                    buf[cbuf++] = cmp_header->byte[3];          // ̂܂8bit
                    s4=1, s4d = byte & 0x0f;                    // 4bit
                    break;
                case 0x9:       // 9X : (X+2)oCgO2oCgRs[BȂB
#ifdef _DEBUG
            printf("9 ");
#endif
                    if (cbuf + 2 > buf_size){                   // I[o[t[
                        error->code = ERROR_OVERFLOW;
                        error->cbuf = cbuf;
                        error->type = 0x09;
                        error->param = 2;
                        return -1;
                    }
                    dt = (byte & 0x0f) + 2;                     // ʒuvZ
                    if (cbuf<dt || dt<2){                       // ʒus
                        error->code = ERROR_ILG_POS;
                        error->cbuf = cbuf;
                        error->type = 0x09;
                        error->param = dt;
                        return -1;
                    }
                    *(unsigned short*)(buf+cbuf) = *(unsigned short*)(buf+cbuf-dt);
                    cbuf += 2;                                  // 16bitRs[
                    break;
                case 0xA:       // AXYY : {(X+3)+YY}oCgO(X+3)oCgRs[BȂB
#ifdef _DEBUG
            printf("A ");
#endif
                    sz = (byte & 0x0f) + 3;                     // TCYvZ
                    if (cbuf + sz > buf_size){                  // I[o[t[
                        error->code = ERROR_OVERFLOW;
                        error->cbuf = cbuf;
                        error->type = 0x0A;
                        error->param = sz;
                        return -1;
                    }
                    byte = fgetc(file);                         // 1byteǂ(YY)
                    dt = sz + byte;                             // ʒum
                    if (cbuf<dt){                               // ʒus(dt<sz̃`FbN͕sv)
                        error->code = ERROR_ILG_POS;
                        error->cbuf = cbuf;
                        error->type = 0x0A;
                        error->param = dt;
                        return -1;
                    }
                    memcpy(buf+cbuf, buf+cbuf-dt, sz);          // Rs[s
                    cbuf += sz;
                    break;
                case 0xB:       // BXYYZ : (YYZ+103h)oCgO(X+3)oCgRs[BB
#ifdef _DEBUG
            printf("B ");
#endif
                    sz = (byte & 0x0f) + 3;                     // TCYvZ
                    if (cbuf + sz > buf_size){                  // I[o[t[
                        error->code = ERROR_OVERFLOW;
                        error->cbuf = cbuf;
                        error->type = 0x0B;
                        error->param = sz;
                        return -1;
                    }
                    dt = fgetc(file);                           // 1byteǂ(YY)
                    byte = fgetc(file);                         // 1byteǂ(Z)
                    dt = ((dt<<4)|((byte&0xf0)>>4))+0x103;      // ʒuvZ
                    if (cbuf<dt || sz>dt){                      // ʒu܂̓TCYs
                        error->code = ERROR_ILG_POS;
                        error->cbuf = cbuf;
                        error->type = 0x0B;
                        error->param = dt;
                        return -1;
                    }
                    memcpy(buf+cbuf, buf+cbuf-dt, sz);          // Rs[s
                    cbuf += sz;
                    s4=1, s4d = byte & 0x0f;                    // 4bit
                    break;
                case 0xC:       // CXYYZZ : YYZZ(BigEndian)oCgO(X+3)oCgRs[BȂB
#ifdef _DEBUG
            printf("C ");
#endif
                    sz = (byte & 0x0f) + 3;                     // TCYvZ
                    if (cbuf + sz > buf_size){                  // I[o[t[
                        error->code = ERROR_OVERFLOW;
                        error->cbuf = cbuf;
                        error->type = 0x0C;
                        error->param = sz;
                        return -1;
                    }
                    dt = fgetc(file); byte = fgetc(file);       // 2byteǂ(YY, ZZ)
                    dt = (dt << 8) | byte;                      // BigEndianŌ
                    if (cbuf<dt || sz>dt){                      // ʒu܂̓TCYs
                        error->code = ERROR_ILG_POS;
                        error->cbuf = cbuf;
                        error->type = 0x0C;
                        error->param = dt;
                        return -1;
                    }
                    memcpy(buf+cbuf, buf+cbuf-dt, sz);          // Rs[s
                    cbuf += sz;
                    break;
                case 0xD:       // D : 1byteO1byteRs[BB
#ifdef _DEBUG
            printf("D ");
#endif
                    if (cbuf + 1 > buf_size){                   // I[o[t[
                        error->code = ERROR_OVERFLOW;
                        error->cbuf = cbuf;
                        error->type = 0x0D;
                        error->param = 1;
                        return -1;
                    }
                    if (cbuf == 0){                             // ʒus
                        error->code = ERROR_ILG_POS;
                        error->cbuf = cbuf;
                        error->type = 0x0D;
                        error->param = 1;
                        return -1;
                    }
                    buf[cbuf] = buf[cbuf-1];                    // 8bitRs[
                    cbuf++; s4=1, s4d = byte & 0x0f;            // 4bit
                    break;
                case 0xE:       // E : 2byteO2byteRs[BB
#ifdef _DEBUG
            printf("E ");
#endif
                    if (cbuf + 2 > buf_size){                   // I[o[t[
                        error->code = ERROR_OVERFLOW;
                        error->cbuf = cbuf;
                        error->type = 0x0E;
                        error->param = 2;
                        return -1;
                    }
                    if (cbuf < 2){                              // ʒus
                        error->code = ERROR_ILG_POS;
                        error->cbuf = cbuf;
                        error->type = 0x0E;
                        error->param = 2;
                        return -1;
                    }
                    *(unsigned short*)(buf+cbuf) = *(unsigned short*)(buf+cbuf-2);
                    cbuf+=2; s4=1, s4d = byte & 0x0f;           // 16bitRs[, 4bit
                    break;
                case 0xF:       // FX : cmp_header->word[X+1] (16bit)݁BȂB
#ifdef _DEBUG
            printf("F ");
#endif
                    if (cbuf + 2 > buf_size){                   // I[o[t[
                        error->code = ERROR_OVERFLOW;
                        error->cbuf = cbuf;
                        error->type = 0x0F;
                        error->param = 2;
                        return -1;
                    }
                    dt = (byte & 0x0f) + 1;                     // ԍ擾
                    *(unsigned short*)(buf+cbuf) = cmp_header->word[dt];
                    cbuf += 2;                                  // 16bitRs[
                    break;
            }
        }
    }
    
    return -1;
}


void show_error_expansion(const unsigned char *buf, const ErrorInfo *error){
    int i, j;
    printf("Failed to deploy data.\n");
    printf("Expand the location data: %#06x\n", error->cbuf);
    printf("Cause of the error: ");
    switch (error->code){
        case ERROR_OVERFLOW:
            printf("Write overflow %#06x bytes.\n", error->param);
            break;
        case ERROR_ILG_POS:
            printf("Incorrect location or size %#06x bytes before I try to copy.\n", error->param);
            break;
        case ERROR_EOF:
            printf("Unexpected EOF has been detected.");
            break;
        default:
            printf("Unknown");
            break;
    }
    printf("Dump the data to expand the error:\n");
    for (j=0; ; j+=16){
        printf("%04X ", j);
        for (i=0; i<16; i++){
            if (j+i >= error->cbuf) return;
            printf(" %02X", (const unsigned char)buf[j+i]);
        }
        printf("\n");
    }
    return;
}


// SDK`FMF`ɕϊB߂l0(), -1(s)
int convert_to_fmf(FMFHeader *fmf_header, unsigned short *buf_fmf,
const unsigned char *buf_sdk, unsigned int size, int vert, int height)
{
    int length = (size / height)>>1;      // height}X(2*heightoCg)1
    int x, y, i;
    int x_size = vert ? height : length;
    int y_size = vert ? length : height;
    unsigned int chip, hflip, vflip;
    
    // FMF wb_[
    fmf_header->identifier[0] = 'F';
    fmf_header->identifier[1] = 'M';
    fmf_header->identifier[2] = 'F';
    fmf_header->identifier[3] = '_';
    fmf_header->size          = size;
    fmf_header->mapWidth      = x_size;
    fmf_header->mapHeight     = y_size;
    fmf_header->chipWidth     = 32;
    fmf_header->chipHeight    = 32;
    fmf_header->layerCount    = 1;
    fmf_header->bitCount      = 16;
    
    // C[f[^
    for (y=0; y<y_size; y++){
        for (x=0; x<x_size; x++){
            chip = ((const unsigned short*)buf_sdk)[vert ? x + y*height : x*height + y];
            hflip = chip & 0x4000;
            vflip = chip & 0x8000;
            chip &= 0x3fff;
            chip = ((chip>>4)<<8) + (chip&0xf) + ((hflip!=0)*16) + ((vflip!=0)*32);
            ((unsigned short*)buf_fmf)[x + y*x_size] = chip;
        }
    }
    return 0;
}










