
///////////////////////////////////////////////////////////////////////////////
//
//	map_import.cpp
//

#include "map_import.h"

using namespace std;

int LoadIniFileEx(MapGroup *map_groups, int nMaxCount, int *pEndOffset);

int main(int argc, char *argv[]){

	int map_group_count;
	MapGroup map_group[64];
	int map_group_end_offset;

	int size_rom;
	char *buf_rom;
	int size_sdk;
	char *buf_sdk;
	int map_group_num;
	int size_max;
	ifstream ifs;
	ofstream ofs;

	cout << "SDK2 Map Importer\n" << endl;

	// t@C̓ǂݍ
	map_group_count = LoadIniFileEx(map_group, 64, &map_group_end_offset);

	if (argc < 4){
		cout << "Usage: " << argv[0] << " [Name of ROM file] [Compressed map file] [Level number]\n\n""Invalid selection, select from the following:\n" << endl;
		for (int i=0; i<map_group_count; i++){
			cout << i << ": " << map_group[i].name << endl;
		}
		return 0;
	}

	// n`O[vԍ
	map_group_num = atoi(argv[3]);
    if (map_group_num < 0 || map_group_num >= map_group_count){
        cout << "Invalid Selection. Valid numbers are between 0 and " << (map_group_count-1) << "." << endl;
        return -1;
    }

	// ROMt@C̓ǂݍ
	ifs.open(argv[1], ios_base::in | ios_base::binary);
	if (ifs.fail()){
		cout << "Failed to open the ROM file." << endl;
		return 0;
	}
	ifs.seekg(0, ios_base::end);
	size_rom = ifs.tellg();
	ifs.seekg(0, ios_base::beg);
	size_rom -= ifs.tellg();
    buf_rom = new char[size_rom];
    ifs.read(buf_rom, size_rom);
    ifs.close();

	// SDKt@C̓ǂݍ
	ifs.open(argv[2], ios_base::in | ios_base::binary);
	if (ifs.fail()){
		cout << "Failed to open the compressed map file." << endl;
		delete[] buf_rom;
		return 0;
	}
	ifs.seekg(0, ios_base::end);
	size_sdk = ifs.tellg();
	ifs.seekg(0, ios_base::beg);
	size_sdk -= ifs.tellg();
	cout << "Compressed map file size: " << size_sdk << "bytes." << endl;
	buf_sdk = new char[size_sdk];
	ifs.read(buf_sdk, size_sdk);
	ifs.close();

	// ROM̃obt@ɏ
	size_max = ((map_group_num+1 == map_group_count) ? map_group_end_offset : map_group[map_group_num+1].offset)
		- map_group[map_group_num].offset;
	cout << "Available ROM data size: " << size_max << "bytes." << endl;
	if (size_sdk > size_max){
		cout << "Compressed map file is too large. Try removing some unnecessary data." << endl;
		delete[] buf_rom;
		delete[] buf_sdk;
		return 0;
	}
	memcpy_s(buf_rom + map_group[map_group_num].offset, size_max, buf_sdk, size_sdk);

	// ROMt@Co
	ofs.open(argv[1], ios_base::out | ios_base::trunc | ios_base::binary);
	if (ofs.fail()){
		cout << "Failed to save the ROM file." << endl;
		delete[] buf_rom;
		delete[] buf_sdk;
		return 0;
	}
	ofs.write(buf_rom, size_rom);
	ofs.close();

	cout << "Successfully written data to the ROM file." << endl;

	delete[] buf_rom;
	delete[] buf_sdk;
	return 0;
}




// t@C̓ǂݍ
int LoadIniFileEx(MapGroup *map_groups, int nMaxCount, int *pEndOffset){

    char IniFileName[260];
	if (!GetModuleFileName(NULL, IniFileName, sizeof IniFileName)) return -1;
	char *last_sep = strrchr(IniFileName, '\\');
	if (last_sep == NULL) return -1;
	strcpy_s(last_sep+1, ((IniFileName + (sizeof IniFileName)) - last_sep) - 1, "map_converter.ini");

    int count = GetPrivateProfileInt("MapGroups", "Count", 15, IniFileName);
	*pEndOffset = GetPrivateProfileInt("MapGroups", "End_Offset", 0, IniFileName);
    if (count > nMaxCount){
        count = nMaxCount;
    }
    for (int i=0; i<count; i++){
        char str[4];
        sprintf_s(str, sizeof str, "%d", i);
        GetPrivateProfileString("MapGroupName", str, "?",
        map_groups[i].name, sizeof(map_groups[i].name), IniFileName);
        map_groups[i].offset    = GetPrivateProfileInt("MapGroupOffset", str, 0, IniFileName);
        map_groups[i].vertical  = GetPrivateProfileInt("MapGroupVertical", str, 0, IniFileName);
        map_groups[i].height    = GetPrivateProfileInt("MapGroupHeight", str, 16, IniFileName);
    }
    return count;
}








